/* songPlayerControls.C

   Copyright (C) 1999 by David Rose.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

   To contact the original author of this program, write to:
           David Rose
           224 W. Dryden St Apt. 418
	   Glendale, CA 91202
	   USA

	   drose@lafn.org
   */

#include "system.h"

#include "songPlayerControls.h"
#include "tableInfo.h"
#include "tableDataProxy.h"
#include "dataCell.h"
#include "cdReader.h"
#include "paloma.h"
#include "volume.h"

#include "rew.xbm"
#include "stop.xbm"
#include "play.xbm"

#include <unistd.h>

// These Gdk_Pixmaps and Bitmaps are declared static, in a feeble
// attempt to minimize redundant pixmaps on the X server.  Hopefully
// this way only one copy of each pixmap will be created for the
// application, and this one copy will be reused for each window.
// (Although this can't work for the insensitive pixmaps, which get
// built on-the-fly.  Hmm...)

// I don't declare the Gtk_Pixmaps statically, because those are
// widgets in the normal sense, and get added to their parent widgets.
// I'm not sure of the implications of adding a single widget to
// multiple parents.

Gdk_Pixmap SongPlayerControls::_rewind_gdk_pixmap;
Gdk_Bitmap SongPlayerControls::_rewind_gdk_mask;
Gdk_Pixmap SongPlayerControls::_stop_gdk_pixmap;
Gdk_Bitmap SongPlayerControls::_stop_gdk_mask;
Gdk_Pixmap SongPlayerControls::_play_gdk_pixmap;
Gdk_Bitmap SongPlayerControls::_play_gdk_mask;
bool SongPlayerControls::_got_pixmaps = false;


SongPlayerControls::
SongPlayerControls(const Dataset &data) : 
  _data(data),
  _play_adjustment(0.0, 0.0, 100.0, 1.0, 10.0, 0.0),
  _play_position(_play_adjustment)
{
  _internal_update = false;
  _playing = false;
  _current_song_index = -1;
  _length = 0.0;

  make_pixmaps();

  _rewind_button.add(_rewind_pixmap);
  _rewind_pixmap.show();
  _rewind_pixmap.set_build_insensitive(true);

  _stop_button.add(_stop_pixmap);
  _stop_pixmap.show();
  _stop_pixmap.set_build_insensitive(true);

  _play_button.add(_play_pixmap);
  _play_pixmap.show();
  _play_pixmap.set_build_insensitive(true);

  layout_window();
  set_overall_insensitive(true);

  // Now connect up the client events.
  connect_to_method(_client.update_time, this, &got_update_time);
  connect_to_method(_client.ready, this, &got_ready);
  connect_to_method(_client.start_playing, this, &got_start_playing);
  connect_to_method(_client.interrupt, this, &got_interrupt);
  connect_to_method(_client.rstart_state, this, &got_rstart_state);
}

SongPlayerControls::
~SongPlayerControls() {
}

void SongPlayerControls::
set_song(TableDataProxy song, CDReader *cd_reader, 
	 int song_index) {
  assert(song.get_table_id() == TI_songs);
  _current_song_index = song_index;

  _client.clear();

  int repeat_count = 0;

  while (true) {
    string rel_filename = song[SFI_filename].get_string();
    Volume volume(_data, song[SFI_volume_id].get_foreign(TI_volumes));
    if (volume.is_valid()) {
      _filename = volume.get_filename(rel_filename);
    } else {
      _filename = "";
    }
    _length = song[SFI_length].get_float();

    bool good_file = false;
    if (_filename.empty()) {
      // If we don't have a filename, we can play it only if we happen
      // to have a CDReader handy.
#ifdef HAVE_CDAUDIO
      if (cd_reader != NULL) {
	_length = cd_reader->get_track_length(song_index);
	_client.set_cd_track(cd_reader->get_device_name(),
			     song_index + 1, _length);
	good_file = true;
      }
#endif
    } else {
      // Does the file exist and can we read it?
      if (access(_filename.c_str(), R_OK) == 0) {
	_client.set_filename(_filename, song[SFI_file_type].get_int(-1), _length);
	good_file = true;
      }
    }
    
    if (good_file) {
      set_overall_insensitive(true);
      rewind();
      return;
    }

    // We weren't able to load the file.  Should we give up?
    if (repeat_count > 0) {
      _client.stop();
      _client.clear();
      _playing = false;
      return;
    }

    // Try once more.  Maybe the file has just moved.  Query the
    // database again to get the latest song definition.
    song = TableDataProxy(TI_songs, song.get_primary_key());
    repeat_count++;
  }
}

int SongPlayerControls::
get_current_song() const {
  return _current_song_index;
}
 
void SongPlayerControls::
rewind() {
  if (_client.is_playing()) {
    _client.stop();
    _playing = false;
    _client.play(0.0);
  } else {
    _internal_update = true;
    _play_adjustment.set_value(0.0);
  }
}
 
void SongPlayerControls::
stop() {
  _client.stop();
  _playing = false;
}

void SongPlayerControls::
play() {
  if (!_client.is_playing()) {
    double start_pos = _play_adjustment.get_value() / 100.0;
    if (start_pos > 0.9999) {
      start_pos = 0.0;
    }
    _client.play(start_pos * _length);
  }
}

bool SongPlayerControls::
is_playing() const {
  return _playing;
}

Signal0 &SongPlayerControls::
get_end_signal() {
  return _client.end;
}

void SongPlayerControls::
got_ready() {
  if (_playing) {
    // If we thought we were already playing when we got a "ready",
    // that means we must have just switched songs, and we should
    // restart.
    _client.play(0.0);
  }
  set_overall_insensitive(false);
}

void SongPlayerControls::
got_start_playing() {
  _playing = true;
}

void SongPlayerControls::
got_interrupt() {
  _playing = false;
}

void SongPlayerControls::
got_update_time(double time) {
  if (_playing && _length != 0.0) {
    // Set this flag so user_update_position() won't be fooled.
    _internal_update = true;
    _play_adjustment.set_value(time / _length * 100.0);
  }
}

void SongPlayerControls::
got_rstart_state(bool flag) {
  set_slider_insensitive(!flag);
}

void SongPlayerControls::
user_update_position() {
  if (!_internal_update && _client.is_playing()) {
    // The user must have grabbed the slider while the music was
    // playing.  Better stop the music, or things will just get
    // confused.
    _client.stop();
    _playing = false;
  }
  _internal_update = false;
}
 
void SongPlayerControls::
rewind_button_clicked() {
  rewind();
}
 
void SongPlayerControls::
stop_button_clicked() {
  stop();
}

void SongPlayerControls::
play_button_clicked() {
  play();
}

void SongPlayerControls::
make_pixmaps() {
  Gdk_Window window = paloma->get_window();
  if (!_got_pixmaps) {
    
    // What a lot of work just to create a viewable bitmap image.
    _rewind_gdk_pixmap.create_from_data
      (window, (const gchar *)rew_bits, 
       rew_width, rew_height, window.get_depth(), 
       get_colormap().black(), get_colormap().white());
    _rewind_gdk_mask.create
      (window, (gchar *)rew_bits, rew_width, rew_height);

    _stop_gdk_pixmap.create_from_data
      (window, (const gchar *)stop_bits, 
       stop_width, stop_height, window.get_depth(), 
       get_colormap().black(), get_colormap().white());
    _stop_gdk_mask.create
      (window, (gchar *)stop_bits, stop_width, stop_height);

    _play_gdk_pixmap.create_from_data
      (window, (const gchar *)play_bits, 
       play_width, play_height, window.get_depth(), 
       get_colormap().black(), get_colormap().white());
    _play_gdk_mask.create
      (window, (gchar *)play_bits, play_width, play_height);

    _got_pixmaps = true;
  }

  _rewind_pixmap.set(_rewind_gdk_pixmap, _rewind_gdk_mask);
  _stop_pixmap.set(_stop_gdk_pixmap, _stop_gdk_mask);
  _play_pixmap.set(_play_gdk_pixmap, _play_gdk_mask);
}

void SongPlayerControls::
layout_window() {
  //  set_shadow_type(GTK_SHADOW_IN);
  set_border_width(4);

  // An outer box just to supply some spacing.
  add(_box0);
  _box0.show();
  _box0.set_border_width(8);

  // The main vbox.
  _box0.pack_start(_box1, true, true, 0);
  _box1.set_spacing(8);
  _box1.show();

  // The position scale.
  _play_position.set_draw_value(false);
  _play_position.set_update_policy(GTK_UPDATE_CONTINUOUS);
  _box1.pack_start(_play_position, true, true, 0);
  _play_position.show();

  connect_to_method(_play_adjustment.value_changed, this, 
		    &user_update_position);

  // A box for the rewind/stop/play buttons.
  _box1.pack_start(_box2, true, true, 0);
  _box2.set_spacing(8);
  _box2.show();

  // The Rewind button.
  _rewind_button.set_usize(60, 32);
  _box2.pack_start(_rewind_button, true, true, 0);
  _rewind_button.show();
  connect_to_method(_rewind_button.clicked, this, &rewind_button_clicked);

  // The Stop button.
  _stop_button.set_usize(60, 32);
  _box2.pack_start(_stop_button, true, true, 0);
  _stop_button.show();
  connect_to_method(_stop_button.clicked, this, &stop_button_clicked);

  // The Play button.
  _play_button.set_usize(60, 32);
  _box2.pack_start(_play_button, true, true, 0);
  _play_button.show();
  connect_to_method(_play_button.clicked, this, &play_button_clicked);
}

void SongPlayerControls::
set_overall_insensitive(bool flag) {
  set_sensitive(!flag);
}

void SongPlayerControls::
set_slider_insensitive(bool flag) {
  _play_position.set_sensitive(!flag);
}
